
class ModelRegistery:

	METADATA = {

		'car_v1' : {
			'model_name' : 'car_v1',
			'description' : 'In this model the car have 5 sensors and a discret action space',
			'observation_space' : {'name' : 'basic_sensor',
										'observer_type' : 'Camera',
										'dimension' : (5,),
										'values_type' : 'float',
										'values_range' : [0, 20]
			},
			'action_space' : {'descritpion' : 'Discret orientation and discret acceleration',
								'type': 'discret',
								'dimension' : (3,),	
								'actioner_names' : ['accelerator', 'breaks', 'wheel_orientation'],
								'actioner_types' : {'accelerator': 'int', 'breaks': 'int', 'wheel_orientation': 'int'},
								'actioner_ranges' :  {'accelerator': {0, 1, 2}, 'breaks': {0, 1, 2}, 'wheel_orientation':{-50, -20, 0, 20, 50}}

			}
		},

		'car_v2' : {
			'model_name' : 'car_v2',
			'description' : 'In this model the car have 5 sensors and a continuous action space',
			'observation_space' : {'name' : 'basic_sensor',
										'observer_type' : 'Camera',
										'dimension' : (5,),
										'values_type' : 'float',
										'values_range' : [0, 20]
			},
			'action_space' : {'descritpion' : 'Discret orientation and discret acceleration',
								'type': 'discret',
								'dimension' : (3,),	
								'actioner_names' : ['accelerator', 'breaks', 'wheel_orientation'],
								'actioner_types' : {'accelerator': 'float', 'breaks': 'float', 'wheel_orientation': 'float'},
								'actioner_ranges' :  {'accelerator': [0, 1], 'breaks': [0, 1], 'wheel_orientation':[-90, 90]}


			}
		}

	}



from pyrace.model import Model
from pyrace.vue import Vue
from pyrace.setting import Setting
from pyrace.model.modelregistery import ModelRegistery
from pyrace.vue.mapregistery import MapRegistery
import matplotlib.pyplot as plt
import numpy
import pprint
pp = pprint.PrettyPrinter(indent=4)



class Env:
    def __init__(self, model_name=''):

        self.vue = Vue(Setting.MAP_INIT)
        try :
            self.model = Model(model_name, self.vue)
            self.model_name = model_name
        except KeyError :
            print('Please enter a valide model name. Select one of the following model en reload Environment\n')
            self.get_model_list()
            self.model = Model(Setting.MODEL_INIT, self.vue)
            self.model_name = Setting.MODEL_INIT
        except :
            raise 

        self.action_space = self.model.car.action_space
        self.observation_space = self.model.car.observation_space

     

    def load(self, new_map_name=''):
        if new_map_name in list(MapRegistery.METADATA.keys()) :
            self.vue = Vue(new_map_name)
            print ('Map {} loaded and a new enviroment ready !'.format(new_map_name))
            _ = self.reset()
        else :
            print('Please type a valide map name : \n')
            self.get_map_list()
        return None


    def sample_action(self):
        return self.model.car.action_space.sample()


    def step(self, action):
        output = self.model.step(action)
        return output

    def render(self):
        if not plt.fignum_exists(1) :
            self.fig, self.ax = plt.subplots()
            self.render_img = self.ax.imshow(self.vue.map_img)

        built_img = self.vue.trace(self.model.car)
        self.render_img.set_data(built_img) 
        self.fig.canvas.draw_idle()
        plt.axis('off')
        plt.pause(1/Setting.FRAME_FREQ)


    def reset(self):
        plt.close()
        self.model = Model(self.model_name, self.vue)
        self.action_space = self.model.car.action_space
        self.observation_space = self.model.car.observation_space
        return self.model.state


    def get_model_info(self):
        pp.pprint(ModelRegistery.METADATA[self.model_name])


    def get_model_list(self):
        print(ModelRegistery.METADATA.keys())

    def get_map_list(self):
        print(MapRegistery.METADATA.keys())






from pyrace.setting import Setting
from pyrace.model.modelregistery import ModelRegistery
from pyrace.model.car import Car

class Model:
    def __init__(self, model_name, vue):

        self.metadata = ModelRegistery.METADATA[model_name]

        self.vue = vue
        self.car = Car(self.vue.metadata, self.metadata)
        
        self.progress = 0
        
        self.state = self.car.get_state(self.vue)
        self.reward = 0
        self.done = False
        self.info = {}
        self.history = [(self.state, self.reward, self.done, self.info)]


    def step(self, action):
        
        if self.car.step(action) :

            self.state = self.car.get_state(self.vue)

            progress = self.vue.get_progress(self.car.position)
            self.reward = progress - self.progress
            if self.reward > 1000 :
                self.reward = -self.reward  # avoid going directly back during after start. 
                self.done = True

            self.progress = progress

            if not self.vue.is_road(self.car.position):
                self.done = True

            self.info = {}

            self.history.append((self.state, self.reward, self.done, self.info))
            return self.state, self.reward, self.done, self.info

        else :
            return None







    




import pathlib


class Setting:
    MAP_INIT = 'map_1'
    MODEL_INIT = 'car_v1'

    CWD = pathlib.Path(__file__).parent

    CAR_IMG_PATH = CWD / "data/car_img.png"
    MARKER_IMG_PATH = CWD / "data/marker_img.png"

    MARKER_SIZE = 16

    FRAME_FREQ = 5

    ACCELERATIOR_SENSITIVITY = 1
    BREAKS_SENSITIVITY = 1

    TIC = 1




from pyrace.model.space import Space
from pyrace.model.camera import Camera
from pyrace.setting import Setting
import numpy

import pprint
pp = pprint.PrettyPrinter(indent=4)



class Car:
    def __init__(self, vue_metadata, model_metadata):
        self.model_metadata = model_metadata

        self.action_space = Space(model_metadata, 'action')
        self.observation_space = Space(model_metadata, 'observation')

        self.position = vue_metadata['init']['position']
        self.orientation = vue_metadata['init']['orientation']

        self.speed = numpy.array([0, 0])
        self.acceleration = numpy.array([0, 0])

        self.car_actioners = {'accelerator' : 0, 'breaks' : 0, 'wheel_orientation' : 0}

        self.accelerator_sensitivity = Setting.ACCELERATIOR_SENSITIVITY
        self.breaks_sensitivity = Setting.BREAKS_SENSITIVITY


        self.observer = self.init_observer()


    def init_observer(self):
        observer_description = self.observation_space.description

        if observer_description['observer_type'] == 'Camera' :
            return Camera(observer_description)

        elif observer_description['observers_type'] == 'GPS':
            print('Coming soon !')
            return None

        elif observer_description['observers_type'] == '360_Sensor':
            print('Coming soon !')
            return None



    def get_state(self, vue):
        state = self.observer.get_value(vue, self.position, self.orientation)
        return state


    def step(self, action):
        if self.action_space.check(action):
            idx = 0
            for actioner in self.car_actioners.keys() :
                if actioner in self.action_space.description['actioner_names'] :
                    self.car_actioners[actioner] = action[idx]
                    idx += 1

            # update acceleration
            alpha = self.car_actioners['wheel_orientation']
            alpha_rad = numpy.pi*alpha/180
            accelerator_effect = self.car_actioners['accelerator']*numpy.array([-numpy.sin(alpha_rad), numpy.cos(alpha_rad)])
            breaks_effect = self.car_actioners['breaks']*numpy.array([numpy.sin(alpha_rad), -numpy.cos(alpha_rad)])
            
            self.acceleration = self.accelerator_sensitivity*accelerator_effect + self.breaks_sensitivity*breaks_effect

            # update speed
            self.speed = self.speed + Setting.TIC*self.acceleration

            # from car referential to map referential
            theta = self.orientation
            theta_rad = numpy.pi*theta/180
            rot_mat = numpy.array([[numpy.cos(theta_rad), -numpy.sin(theta_rad)], [-numpy.sin(theta_rad), -numpy.cos(theta_rad)]])
            speedxy = numpy.matmul(rot_mat, self.speed)

            # Drift effect pending dev !

            # update position and orientation
            self.position = self.position + Setting.TIC*speedxy
            self.orientation = alpha + theta

            return True
        else :
            print('Input action does not match action space : \n')
            pp.pprint(self.action_space.description)
            return False





